InterpreterPlugin subclass: #DirectoryPlugin
	instanceVariableNames: 'hasSecurityPlugin smallIntegerMaxVal '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins'!

!DirectoryPlugin commentStamp: 'dtl 2/12/2002 19:02' prior: 0!
I provide access to a Posix directory stream for efficient access to directory contents. I also know how to read file status and how to create a DirectoryEntry for a file. Methods for primitive access and other utilities are provided in my class methods. Class methods in the "directory access" and "file access" categories should be used rather directly calling the primitives where possible.

The directory stream primitives use the address of a directory stream as a handle to identify a directory stream. This permits several directory streams to be opened and manipulated independently by separate Smalltalk processes. However, care must be taken to avoid passing the handle for an invalid directory stream (such as one which has already been closed) to one of these methods.

The plugin has been tested on Linux and on Windows. The directory stream support produces a performance improvement only on Unix (Linux) systems. The file access primitives provide substantial improvements for both Unix and Windows systems. Other platforms have not been tested.

Hooks are placed in various methods in the FileStream and FileDirectory hierarchies where measurable performance improvements can be seen for both Unix and Windows. The number of these changes is kept as small as possible.

Security checks for this plugin are implemented but have not been tested. It is likely that using the security plugin will reduce the performance of this plugin significantly.

The semantics of #isAFileNamed: are different as a result of this change set. On Unix, the new method (implemented on the class side of the plugin) answers false if the specified path is a directory, whereas the default method incorrectly answered true. Other differences may exist on Windows. The new implementation is reasonable and hopefully less incorrect. However it could break some existing applications.!


!DirectoryPlugin methodsFor: 'directory primitives' stamp: 'dtl 2/3/2002 10:17'!
primitiveClosedir
	"Close the directory stream for dirPointerOop. Answer dirPointerOop on success."

	| dirPointerOop dirStream result |
	self export: true.
	self var: 'dirStream' declareC: 'DIR *dirStream'.
	dirPointerOop _ interpreterProxy stackObjectValue: 0.
	(dirPointerOop = interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(interpreterProxy is: dirPointerOop KindOf: 'ExternalAddress') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveClosedir: invalid argument, expected an ExternalAddress\n")'.
		^ interpreterProxy primitiveFail].
	dirStream _ self pointerFrom: dirPointerOop.
	(dirStream = nil) ifTrue: [^ interpreterProxy primitiveFail].
	result _ self closedir: dirStream.
	(result = -1)
		ifTrue: [^ interpreterProxy primitiveFail]
		ifFalse: [interpreterProxy pop: 2; push: dirPointerOop]
! !

!DirectoryPlugin methodsFor: 'directory primitives' stamp: 'dtl 2/3/2002 10:18'!
primitiveOpendir

	"DirectoryPlugin primOpendir: '/etc'"

	| dirName dir dirOop |
	self export: true.
	self var: 'dir' declareC: 'DIR *dir'.
	self var: 'dirOopArrayPointer' declareC: 'unsigned char *dirOopArrayPointer'.
	dirName _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: dirName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveOpendir: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	(dir _ self openDirectoryStream: dirName) isNil ifTrue: [^ interpreterProxy primitiveFail].
	dirOop _ self addressObjectFor: dir.
	interpreterProxy pop: 2; push: dirOop
! !

!DirectoryPlugin methodsFor: 'directory primitives' stamp: 'dtl 2/3/2002 10:16'!
primitiveReaddir
	"Get the next entry in the directory stream. Answer the name of the entry, or
	error for end of directory stream."

	| dirPointerOop dirStream ent entryName |
	self export: true.
	self var: 'ent' declareC: 'struct dirent *ent'.
	self var: 'dirStream' declareC: 'DIR *dirStream'.
	dirPointerOop _ interpreterProxy stackObjectValue: 0.
	(dirPointerOop = interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(interpreterProxy is: dirPointerOop KindOf: 'ExternalAddress') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveReaddir: invalid argument, expected an ExternalAddress\n")'.
		^ interpreterProxy primitiveFail].
	dirStream _ self pointerFrom: dirPointerOop.
	(dirStream = nil) ifTrue: [^ interpreterProxy primitiveFail].
	ent _ self readdir: (self pointerFrom: dirPointerOop).
	(ent = nil) ifTrue: [^ interpreterProxy primitiveFail].
	entryName _ self stringFromCString: (self cCode: 'ent->d_name' inSmalltalk: [0]).
	interpreterProxy pop: 2; push: entryName
! !

!DirectoryPlugin methodsFor: 'directory primitives' stamp: 'dtl 2/3/2002 10:16'!
primitiveRewinddir
	"Set directoryStream to first entry. Answer dirPointerOop."

	| dirPointerOop dirStream |
	self export: true.
	self var: 'dirStream' declareC: 'DIR *dirStream'.
	dirPointerOop _ interpreterProxy stackObjectValue: 0.
	(dirPointerOop = interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(interpreterProxy is: dirPointerOop KindOf: 'ExternalAddress') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveRewindir: invalid argument, expected an ExternalAddress\n")'.
		^ interpreterProxy primitiveFail].
	dirStream _ self pointerFrom: dirPointerOop.
	(dirStream = nil) ifTrue: [^ interpreterProxy primitiveFail].
	self rewinddir: dirStream.
	interpreterProxy pop: 2; push: dirPointerOop
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:13'!
primitiveFileAccessExecutable
	"Check if file is executable by this OS process."

	| pathString |
	self export: true.
	pathString _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: pathString KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExists: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: pathString mode: self fileExecutableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:15'!
primitiveFileAccessExecutableInDirectory
	"Check if file is executable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	| directoryPath fileName |
	self export: true.
	directoryPath _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: directoryPath KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid directory path argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	fileName _ interpreterProxy stackObjectValue: 1.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid file name argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: fileName directoryPath: directoryPath mode: self fileExecutableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:12'!
primitiveFileAccessExists
	"Check for existence of a file with a call to access()."

	| pathString |
	self export: true.
	pathString _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: pathString KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExists: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: pathString mode: self fileOKFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:14'!
primitiveFileAccessExistsInDirectory
	"Check for existence of a file in a directory with a call to access(). The
	directory path string is expected to include a trailing directory separator,
	for example '/etc/' rather than '/etc'."

	| directoryPath fileName |
	self export: true.
	directoryPath _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: directoryPath KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid directory path argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	fileName _ interpreterProxy stackObjectValue: 1.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid file name argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: fileName directoryPath: directoryPath mode: self fileOKFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:13'!
primitiveFileAccessReadable
	"Check if file is readable by this OS process."

	| pathString |
	self export: true.
	pathString _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: pathString KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExists: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: pathString mode: self fileReadableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:16'!
primitiveFileAccessReadableInDirectory
	"Check if file is readable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	| directoryPath fileName |
	self export: true.
	directoryPath _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: directoryPath KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid directory path argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	fileName _ interpreterProxy stackObjectValue: 1.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid file name argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: fileName directoryPath: directoryPath mode: self fileReadableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:16'!
primitiveFileAccessWritable
	"Check if file is writable by this OS process."

	| pathString |
	self export: true.
	pathString _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: pathString KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExists: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: pathString mode: self fileWritableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/9/2002 14:16'!
primitiveFileAccessWritableInDirectory
	"Check if file is writable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	| directoryPath fileName |
	self export: true.
	directoryPath _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: directoryPath KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid directory path argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	fileName _ interpreterProxy stackObjectValue: 1.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveFileAccessExistsIndirectory: invalid file name argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	((self checkAccess: fileName directoryPath: directoryPath mode: self fileWritableFlag) = 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/3/2002 10:24'!
primitiveStat
	"Stat a file using a path name independent of the current directoryPath"

	| fileName statArray |
	self export: true.
	fileName _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveStat: invalid argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	statArray _ self statFileToEntryArray: fileName.
	(statArray = 0)
		ifTrue: [^ interpreterProxy primitiveFail]
		ifFalse: [interpreterProxy pop: 2; push: statArray]
! !

!DirectoryPlugin methodsFor: 'file primitives' stamp: 'dtl 2/10/2002 16:57'!
primitiveStatInPath
	"Stat a file in a directory path. The directory path string is expected to include
	a trailing directory separator, for example '/etc/' rather than '/etc'."

	| fileName directoryPath statArray |
	self export: true.
	directoryPath _ interpreterProxy stackObjectValue: 0.
	(interpreterProxy is: directoryPath KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveStatInPath: invalid directory path argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	fileName _ interpreterProxy stackObjectValue: 1.
	(interpreterProxy is: fileName KindOf: 'String') ifFalse:
		[self cCode: 'fprintf(stderr, "primitiveStatInPath: invalid file name argument, expected a String\n")'.
		^ interpreterProxy primitiveFail].
	statArray _ self statToEntryArray: fileName directoryPath: directoryPath.
	(statArray = 0)
		ifTrue: [^ interpreterProxy primitiveFail]
		ifFalse: [interpreterProxy pop: 3; push: statArray]
! !

!DirectoryPlugin methodsFor: 'other primitives' stamp: 'dtl 2/3/2002 16:51'!
primitiveModuleIsLoaded
	"Answer true to indicate that the module is running."

	self export: true.
	interpreterProxy pop: 1; push: interpreterProxy trueObject
! !

!DirectoryPlugin methodsFor: 'other primitives' stamp: 'dtl 2/3/2002 16:51'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	self export: true.
	self returnTypeC: 'void'.
	interpreterProxy pop: 1; push: (self stringFromCString: self versionString)
! !

!DirectoryPlugin methodsFor: 'version string' stamp: 'dtl 1/27/2002 10:49'!
versionString
	"Answer a string containing the version string for this plugin."

	| version |
	self returnTypeC: 'char *'.

	self var: 'version' declareC: 'static char version[]= "1.3"'.
	^ self cCode: 'version' inSmalltalk: ['1.3']
! !

!DirectoryPlugin methodsFor: 'private - directory' stamp: 'dtl 2/3/2002 11:32'!
openDirectoryStream: pathOOP
	"Open a new directory stream. Answer a pointer to the directory stream or NULL."

	| sPtr len idx dirNameCString |
	self var: 'sPtr' declareC: 'char *sPtr'.
	self var: 'dirNameCString' declareC: 'char dirNameCString[BUFSIZ]'.
	self var: 'd' declareC: 'static struct dirent *d'.
	self returnTypeC: 'DIR *'.
	sPtr _ interpreterProxy arrayValueOf: pathOOP.
	len _ interpreterProxy sizeOfSTArrayFromCPrimitive: sPtr.
	(self canOpenDirectoryStreamFor: sPtr length: len)
		ifTrue:
			[idx _ 0.
			[idx < len] whileTrue:
				[dirNameCString at: idx put: (sPtr at: idx).
				idx _ idx + 1].
			dirNameCString at: len put: 0.
			^ self opendir: dirNameCString]
		ifFalse:
			[^ nil].
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:50'!
checkAccess: fileName directoryPath: directoryPath mode: mode
	"Check access to fileName in directoryPath. The directory path string should end
	with a path separator (slash)."

	| nameLen dirPathLen cString namePtr dirPathPtr len idx nameIdx |
	self var: 'cString' declareC: 'char cString[1024]'.
	self var: 'namePtr' declareC: 'char *namePtr'.
	self var: 'dirPathPtr' declareC: 'char *dirPathPtr'.
	nameLen _ interpreterProxy stSizeOf: fileName.
	dirPathLen _ interpreterProxy stSizeOf: directoryPath.
	len _ nameLen + dirPathLen.	"Space for a null terminated string"
	"Note: The static sized string buffer is faster than using a Smalltalk allocated
	string as the buffer, and is unlikely to fail in practice. In the event that
	the buffer turns out to be too small, write an error message to stdout before
	failing."
	(len >= 1024)
		ifTrue:
			[self cCode: 'fprintf(stderr, "UnixDirectoryPlugin: path name too long in checkAccessdirectoryPathmode()\n")'.
			^ 0].
	namePtr _ interpreterProxy arrayValueOf: fileName.
	dirPathPtr _ interpreterProxy arrayValueOf: directoryPath.
	"Copy path and file name into a full path string"
	idx _ 0.
	[idx < dirPathLen] whileTrue:
		[cString at: idx put: (dirPathPtr at: idx).
		idx _ idx + 1].
	nameIdx _ 0.
	[nameIdx < nameLen] whileTrue:
		[cString at: idx put: (namePtr at: nameIdx).
		nameIdx _ nameIdx + 1.
		idx _ idx + 1].
	cString at: idx put: 0.	"Null terminate the string"
	^ self cCode: 'access(cString, mode)' inSmalltalk: [0].
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:50'!
checkAccess: pathString mode: mode
	"Check access to pathString."

	| cString len sPtr idx |
	self export: true.
	self var: 'cString' declareC: 'char cString[1024]'.
	self var: 'sPtr' declareC: 'char *sPtr'.
	len _ interpreterProxy stSizeOf: pathString.
	"Note: The static sized string buffer is faster than using a Smalltalk allocated
	string as the buffer, and is unlikely to fail in practice. In the event that
	the buffer turns out to be too small, write an error message to stdout before
	failing."
	(len >= 1024)
		ifTrue:
			[self cCode: 'fprintf(stderr, "UnixDirectoryPlugin: path name too long in checkAccessmode()\n")'.
			^ 0].
	"Copy pathString to the new string"
	sPtr _ interpreterProxy arrayValueOf: pathString.
	((self canStatFilePath: sPtr length: len) = 0) ifTrue: [^ 0].
	idx _ 0.
	[idx < len]
		whileTrue:
			[cString at: idx put: (sPtr at: idx).
			idx _ idx + 1].
	cString at: idx put: 0.
	^ self cCode: 'access(cString, mode)' inSmalltalk: [0].
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/2/2002 10:47'!
entryArrayFor: fileName fromStatBuf: statBuf
	"Answer a file entry array from pathName. If trueFalseFlag is non-zero,
	interpret pathName relative to the directoryPath; otherwise treat it as
	a file path name independent of the directoryPath."

	| creationDate modificationDate isDirectory sizeIfFile |
	self var: 'statBuf' type: 'struct stat *'.
	"Use modification time instead (just like ls)"
	creationDate _ self smalltalkSecondsFromPosixSeconds:
		(self cCode: 'statBuf->st_mtime' inSmalltalk: [0]).
	"Use status change time instead"
	modificationDate _ self smalltalkSecondsFromPosixSeconds:
		(self cCode: 'statBuf->st_ctime' inSmalltalk: [0]).
	((self cCode: 'S_ISDIR(statBuf->st_mode)') = 0)
		ifTrue:
			[isDirectory _ 0.
			sizeIfFile _ self cCode: 'statBuf->st_size']
		ifFalse:
			[isDirectory _ 1.
			sizeIfFile _ 0].
	^ self makeDirEntryName: fileName
		createDate: creationDate
		modDate: modificationDate
		isDir: isDirectory
		fileSize: sizeIfFile.
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:10'!
fileExecutableFlag

	^ self cCode: 'X_OK' inSmalltalk: [1]! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:09'!
fileOKFlag

	^ self cCode: 'F_OK' inSmalltalk: [0]! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:11'!
fileReadableFlag

	^ self cCode: 'R_OK' inSmalltalk: [4]! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/9/2002 14:11'!
fileWritableFlag

	^ self cCode: 'W_OK' inSmalltalk: [2]! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/11/2002 22:54'!
makeDirEntryName: fileName createDate: createDate modDate: modifiedDate isDir: dirFlag fileSize: fileSize
	"Answer an array suitable for creating a DirectoryEntry."

	| nameString createDateOop modifiedDateOop fileSizeOop results |
	self var: 'entryName' declareC: 'char *entryName'.
	interpreterProxy pushRemappableOop: fileName.
	interpreterProxy pushRemappableOop: (self smallOrLargeIntegerObjectFor: createDate).
	interpreterProxy pushRemappableOop: (self smallOrLargeIntegerObjectFor: modifiedDate).
	interpreterProxy pushRemappableOop: (self smallOrLargeIntegerObjectFor: fileSize).
	results _ interpreterProxy instantiateClass: (interpreterProxy classArray) indexableSize: 5.
	fileSizeOop _ interpreterProxy popRemappableOop.
	modifiedDateOop _ interpreterProxy popRemappableOop.
	createDateOop _ interpreterProxy popRemappableOop.
	nameString _ interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 0
		ofObject: results
		withValue: nameString.
	interpreterProxy
		storePointer: 1
		ofObject: results
		withValue: createDateOop.
	interpreterProxy
		storePointer: 2
		ofObject: results
		withValue: modifiedDateOop.
	dirFlag
		ifTrue:
			[interpreterProxy
				storePointer: 3
				ofObject: results
				withValue: interpreterProxy trueObject]
		ifFalse:
			[interpreterProxy
				storePointer: 3
				ofObject: results
				withValue: interpreterProxy falseObject].
	interpreterProxy
		storePointer: 4
		ofObject: results
		withValue: fileSizeOop.
	^ results! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/2/2002 18:08'!
putStatFor: nullTerminatedPathStringPointer intoBuffer: statBufPointer
	"Call stat() on nullTerminatedPathStringPointer, storing the results in
	the buffer at statBufPointer."

	| result |
	self var: 'nullTerminatedPathStringPointer' type: 'char *'.
	self var: 'statBufPointer' type: 'struct stat *'.
	self returnTypeC: 'struct stat *'.
	result _ self
		cCode: '
/* Default is to try to use lstat(). Platforms which do not have it should #define HAVE_LSTAT 0 */
#ifndef HAVE_LSTAT
#define HAVE_LSTAT 1
#endif
#if HAVE_LSTAT
	stat(nullTerminatedPathStringPointer, statBufPointer) && lstat(nullTerminatedPathStringPointer, statBufPointer);
#else
	stat(nullTerminatedPathStringPointer, statBufPointer);
#endif'
		inSmalltalk: [0].
	(result ~= 0)
		ifTrue: [^ 0]
		ifFalse: [^ statBufPointer]
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/3/2002 11:41'!
statFile: pathString intoBuffer: statBuf
	"Call stat() on pathString, putting results in statBuf."

	| sPtr len idx cString |
	self var: 'statBuf' type: 'struct stat *'.
	self var: 'cString' declareC: 'char cString[1024]'.
	self var: 'sPtr' declareC: 'char *sPtr'.
	self returnTypeC: 'struct stat *'.
	len _ interpreterProxy stSizeOf: pathString.
	"Note: The static sized string buffer is faster than using a Smalltalk allocated
	string as the buffer, and is unlikely to fail in practice. In the event that
	the buffer turns out to be too small, write an error message to stdout before
	failing."
	(len >= 1024)
		ifTrue:
			[self cCode: 'fprintf(stderr, "UnixDirectoryPlugin: path name too long in statFileintoBuffer()\n")'.
			^ 0].
	"Copy pathString to the new string"
	sPtr _ interpreterProxy arrayValueOf: pathString.
	((self canStatFilePath: sPtr length: len) = 0) ifTrue: [^ 0].
	idx _ 0.
	[idx < len]
		whileTrue:
			[cString at: idx put: (sPtr at: idx).
			idx _ idx + 1].
	cString at: idx put: 0.
	((self canStatFilePath: cString length: len) = 0)
		ifTrue: [^ 0]
		ifFalse: [^ self putStatFor: cString intoBuffer: statBuf]
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/3/2002 11:40'!
statFile: fileName intoBuffer: statBuf directoryPath: directoryPath
	"Call stat() on fileName in directoryPath, putting results in statBuf. The directory
	path string should end with a path separator (slash)."

	| nameLen dirPathLen cString namePtr dirPathPtr len idx nameIdx |
	self var: 'cString' declareC: 'char cString[1024]'.
	self var: 'statBuf' type: 'struct stat *'.
	self var: 'namePtr' declareC: 'char *namePtr'.
	self var: 'dirPathPtr' declareC: 'char *dirPathPtr'.
	self returnTypeC: 'struct stat *'.
	nameLen _ interpreterProxy stSizeOf: fileName.
	dirPathLen _ interpreterProxy stSizeOf: directoryPath.
	len _ nameLen + dirPathLen.	"Space for a null terminated string"
	"Note: The static sized string buffer is faster than using a Smalltalk allocated
	string as the buffer, and is unlikely to fail in practice. In the event that
	the buffer turns out to be too small, write an error message to stdout before
	failing."
	(len >= 1024)
		ifTrue:
			[self cCode: 'fprintf(stderr, "UnixDirectoryPlugin: path name too long in statFileintoBufferdirectoryPath()\n")'.
			^ 0].
	namePtr _ interpreterProxy arrayValueOf: fileName.
	dirPathPtr _ interpreterProxy arrayValueOf: directoryPath.
	"Copy path and file name into a full path string"
	idx _ 0.
	[idx < dirPathLen] whileTrue:
		[cString at: idx put: (dirPathPtr at: idx).
		idx _ idx + 1].
	nameIdx _ 0.
	[nameIdx < nameLen] whileTrue:
		[cString at: idx put: (namePtr at: nameIdx).
		nameIdx _ nameIdx + 1.
		idx _ idx + 1].
	cString at: idx put: 0.	"Null terminate the string"
	((self canStatFilePath: cString length: len) = 0)
		ifTrue: [^ 0]
		ifFalse: [^ self putStatFor: cString intoBuffer: statBuf]
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/10/2002 14:30'!
statFileToEntryArray: pathName
	"Answer a file entry array from pathName."

	| statBuf statBufPointer |
	self returnTypeC: 'int'.
	self var: 'statBuf' declareC: 'struct stat statBuf'.
	self var: 'statBufPointer' declareC: 'struct stat *statBufPointer= &statBuf'.
	statBufPointer _ self
			statFile: pathName
			intoBuffer: statBufPointer.
	(statBufPointer = 0) ifTrue: [^ 0].
	"Do not know the file name, so pass nil"
	^ self entryArrayFor: interpreterProxy nilObject fromStatBuf: statBufPointer
! !

!DirectoryPlugin methodsFor: 'private - file' stamp: 'dtl 2/2/2002 17:33'!
statToEntryArray: fileName directoryPath: directoryPath
	"Answer a file entry array from pathName."

	| statBuf statBufPointer file |
	self returnTypeC: 'int'.
	self var: 'statBuf' declareC: 'struct stat statBuf'.
	self var: 'statBufPointer' declareC: 'struct stat *statBufPointer'.
	interpreterProxy pushRemappableOop: fileName.
	statBufPointer _ self statFile: fileName
						intoBuffer: (self cCode: '&statBuf' inSmalltalk: [statBuf])
						directoryPath: directoryPath.
	file _ interpreterProxy popRemappableOop.
	(statBufPointer = 0)
		ifTrue: [^ 0]
		ifFalse: [^ self entryArrayFor: file fromStatBuf: statBufPointer]! !

!DirectoryPlugin methodsFor: 'private - security' stamp: 'dtl 2/3/2002 12:03'!
canOpenDirectoryStreamFor: aPathCString length: length
	"Answer non-zero if security permits the directory to be listed."

	"FIXME: This function has not been tested. -dtl"

	| sCLPfn |
	self var: 'aPathCString' type: 'char *'.
	"If the security plugin can be loaded, use it to check . 
	If not, assume it's ok"
	"The hasSecurityPlugin flag is set to 1 by default"
	(hasSecurityPlugin = 0) ifTrue: [^ true].
	sCLPfn _ interpreterProxy ioLoadFunction: 'secCanListPathOfSize' From: 'SecurityPlugin'.
	sCLPfn ~= 0
		ifTrue:
			[^ self cCode: ' ((int (*) (char *, int)) sCLPfn)(aPathCString, length)'
				inSmalltalk: [true]]
		ifFalse:
			[hasSecurityPlugin _ 0.	"Reset the flag so we do not try again next time"
			^ true]
! !

!DirectoryPlugin methodsFor: 'private - security' stamp: 'dtl 2/3/2002 12:02'!
canStatFilePath: aPathCString length: length
	"Answer non-zero if security permits the a stat() call on the file path. Allow a
	stat() call only on files which we are permitted to open."

	"FIXME: This function has not been tested. -dtl"

	| sCOFfn |
	self var: 'aPathCString' type: 'char *'.
	"If the security plugin can be loaded, use it to check . 
	If not, assume it's ok"
	"The hasSecurityPlugin flag is set to 1 by default"
	(hasSecurityPlugin = 0) ifTrue: [^ true].
	sCOFfn _ interpreterProxy ioLoadFunction: 'secCanOpenFileOfSizeWritable' From: 'SecurityPlugin'.
	sCOFfn ~= 0
		ifTrue:
			[^ self cCode: ' ((int (*) (char *, int, int)) sCOFfn)(aPathCString, length, 0)'
				inSmalltalk: [true]]
		ifFalse: 
			[hasSecurityPlugin _ 0.	"Reset the flag so we do not try again next time"
			^ true]
! !

!DirectoryPlugin methodsFor: 'private - timezone' stamp: 'dtl 2/12/2002 17:43'!
smalltalkSecondsFromPosixSeconds: posixSeconds
	"Adapted from Ian's convertToSqueakTime() function in sqXWindow.c. The Smalltalk
	epoch is Jan 1, 1901, and the Posix epoch is Jan 1, 1970. The difference is 17 leap
	years and 52 non-leap years. This method is not tested on Win32."

	| offset tzOffset |
	self var: 'posixSeconds' type: 'time_t'.
	self var: 'offset' declareC: 'time_t offset'.
	self var: 'tzOffset' declareC: 'time_t tzOffset'.
	self returnTypeC: 'time_t'.
	tzOffset _ 0.
	self cCode: '
#ifdef HAVE_TM_GMTOFF
	tzOffset = localtime(&posixSeconds)->tm_gmtoff;
#else
# ifdef HAVE_TIMEZONE
	tzOffset = daylight * 3600 - timezone;
# else
	/* Do something here for Windows */
	tzOffset = _daylight * 3600 - _timezone;
# endif*/
#endif*/'.
	offset _ 2177452800.		"((52 * 365) + (17 * 366)) * 24 * 60 * 60 -> 2177452800"
	^ posixSeconds + offset + tzOffset
! !

!DirectoryPlugin methodsFor: 'private - misc' stamp: 'dtl 1/30/2002 05:39'!
addressObjectFor: aMachineAddress
	"Answer an ExternalAddress object which represents aMachineAddress"

	| addressOop addressOopArrayPointer addressUnion idx |
	self var: 'aMachineAddress' type: 'void *'.
	self var: 'addressUnion'
		declareC: 'union {void *address; unsigned char bytes[sizeof(void *)];} addressUnion'.
	self var: 'addressOopArrayPointer' declareC: 'unsigned char *addressOopArrayPointer'.
	addressOop _ interpreterProxy
		instantiateClass: interpreterProxy classExternalAddress
		indexableSize: self sizeOfPointer.
	addressOopArrayPointer _ interpreterProxy arrayValueOf: addressOop.
	self cCode: 'addressUnion.address = aMachineAddress'.
	idx _ 0.
	[idx < self sizeOfPointer] whileTrue:
		[self cCode: 'addressOopArrayPointer[idx] = addressUnion.bytes[idx]'.
		idx _ idx + 1].
	^ addressOop
! !

!DirectoryPlugin methodsFor: 'private - misc' stamp: 'dtl 2/3/2002 14:56'!
pointerFrom: anExternalAddressOop
	"Answer the machine address contained in anExternalAddressOop."

	| ptr addressUnion idx |
	self returnTypeC: 'void *'.
	self var: 'ptr' declareC: 'unsigned char *ptr'.
	self var: 'addressUnion'
		declareC: 'union {void *address; unsigned char bytes[sizeof(void *)];} addressUnion'.
	((interpreterProxy isBytes: anExternalAddressOop) and:
		[(interpreterProxy stSizeOf: anExternalAddressOop) = self sizeOfPointer])
		ifFalse: [^ nil].
	ptr _ interpreterProxy arrayValueOf: anExternalAddressOop.
	idx _ 0.
	[idx < self sizeOfPointer]	whileTrue:
		[self cCode: 'addressUnion.bytes[idx] = ptr[idx]'.
		idx _ idx + 1].
	^ self cCode: 'addressUnion.address' inSmalltalk: [addressUnion]
! !

!DirectoryPlugin methodsFor: 'private - misc' stamp: 'dtl 1/28/2002 20:54'!
sizeOfPointer
	"Size of a C pointer on this machine"

	^ self cCode: 'sizeof(void *)' inSmalltalk: [4]
! !

!DirectoryPlugin methodsFor: 'private - misc' stamp: 'dtl 2/11/2002 23:08'!
smallOrLargeIntegerObjectFor: aPositiveInteger
	"Answer an Integer object for aPositiveInteger. If the value is too large
	to fit in an SmallInteger, then answer a LargePositiveInteger. Note that
	smallIntegerMaxVal is a class variable with an initialization value determined
	when this class is translated to C code; see my #declareCVarsIn: class method."

	(aPositiveInteger > smallIntegerMaxVal)
		ifTrue: [^ interpreterProxy integerObjectOf: aPositiveInteger]
		ifFalse: [^ interpreterProxy positive32BitIntegerFor: aPositiveInteger]
! !

!DirectoryPlugin methodsFor: 'private - misc' stamp: 'dtl 2/3/2002 17:25'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString stringPtr |
	self var: 'aCString' declareC: 'char *aCString'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	len _ self cCode: 'strlen(aCString)' inSmalltalk: [0].
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	stringPtr _ interpreterProxy arrayValueOf: newString.
	self cCode: '(char *)strncpy(stringPtr, aCString, len)'.
	^ newString

! !


!DirectoryPlugin class methodsFor: 'translation' stamp: 'dtl 8/4/2003 07:43'!
declareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>'.
	cg addHeaderFile: '<dirent.h>'.
	cg addHeaderFile: '<sys/stat.h>'.
	cg addHeaderFile: '<unistd.h>
/* D T Lewis 2002 - ', self moduleName, '.c translated from class ', self name, ' */'.
	"Assume the security plugin can be loaded until proven otherwise"
	cg var: 'hasSecurityPlugin' declareC: 'int hasSecurityPlugin= 1'.
	"The maximum value of a SmallInteger is encoded here so that the plugin can
	be regenerated and rebuilt correctly even if the Smalltalk word size or object
	format changes. -dtl"
	cg var: 'smallIntegerMaxVal' declareC: 'int smallIntegerMaxVal= ', SmallInteger maxVal asString.
! !

!DirectoryPlugin class methodsFor: 'translation' stamp: 'dtl 1/27/2002 10:44'!
moduleName

	^ 'DirectoryPlugin'! !

!DirectoryPlugin class methodsFor: 'translation' stamp: 'dtl 2/11/2002 23:05'!
sdeclareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>'.
	cg addHeaderFile: '<dirent.h>'.
	cg addHeaderFile: '<sys/stat.h>'.
	cg addHeaderFile: '<unistd.h>
/* D T Lewis 2002 - ', self moduleName, '.c translated from class ', self name, ' */'.
	"Assume the security plugin can be loaded until proven otherwise"
	cg var: 'hasSecurityPlugin' declareC: 'static int hasSecurityPlugin= 1'.
	"The maximum value of a SmallInteger is encoded here so that the plugin can
	be regenerated and rebuilt correctly even if the Smalltalk word size or object
	format changes. -dtl"
	cg var: 'smallIntegerMaxVal' declareC: 'static int smallIntegerMaxVal= ', SmallInteger maxVal asString.
! !

!DirectoryPlugin class methodsFor: 'translation' stamp: 'dtl 2/12/2002 17:29'!
shouldBeTranslated
	"Translate one of my subclasses instead."

"	^ false"
	^ true! !

!DirectoryPlugin class methodsFor: 'version testing' stamp: 'dtl 1/27/2002 10:48'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"self installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'version testing' stamp: 'dtl 1/27/2002 10:48'!
versionString
	"Answer version string for this class. If the installed module is up to date, it
	will match this version string."

	"self versionString"

	^ self new versionString! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/12/2002 15:26'!
compareIsAFileNamed: dirPath 
	"Compare the behavior of the old and new versions of #isAFileNamed. 
	Answer an array of all entries for which the old behavior differs from 
	the new behavior."

	"self compareIsAFileNamed: '/etc'"
	"self compareIsAFileNamed: 'C:\WINDOWS'"

	| dir names paths oldTests newTests |
	dir _ FileDirectory on: dirPath.
	names _ dir fileAndDirectoryNames.
	paths _ names collect: [:n | dirPath , FileDirectory slash , n].
	oldTests _ paths collect: [:p | StandardFileStream defaultIsAFileNamed: p].
	newTests _ paths collect: [:p | self isAFileNamed: p].

	"Collect and compare the results"
	^ ((paths
		with: oldTests
		collect: [:n :old | 
			| a | 
			a _ Array new: 3.
			a at: 1 put: n; at: 2 put: old.
			a])
		with: newTests
		collect: [:a :new | 
			a at: 3 put: new.
			a])
		select: [:e | (e at: 2) ~= (e at: 3)]! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/10/2002 14:00'!
timeDirectoryEntryCollection: dirPath

	"self timeDirectoryEntryCollection: '/etc'"
	"self timeDirectoryEntryCollection: 'C:\WINDOWS'"

	| dir oldTime newTime |
	"Evaluate the commands a couple of times to make sure the Unix buffer cache is
	not polluting the timing results."
	dir _ FileDirectory default.
	(dir respondsTo: #defaultDirectoryContentsFor:) ifFalse:
		[^ self inform: 'not used for ', Smalltalk platformName, ', no performance difference on this platform'].
	((dir directoryContentsFor: dirPath) = (dir defaultDirectoryContentsFor: dirPath))
		ifFalse: [self error: 'new method results do not match old method results'].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[100 timesRepeat: [dir defaultDirectoryContentsFor: dirPath]]].
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[100 timesRepeat: [dir directoryContentsFor: dirPath]]].
	^ Array
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/9/2002 22:53'!
timeDirectoryExists: dirPath

	"self timeDirectoryExists: '/etc'"
	"self timeDirectoryExists: 'C:\WINDOWS'"

	| dir names paths oldTime oldTests newTime newTests |
	dir _ FileDirectory on: dirPath.
	names _ dir fileAndDirectoryNames.
	paths _ names collect: [:n | dirPath, FileDirectory slash, n].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[3 timesRepeat: [oldTests _ paths collect: [:p | dir defaultDirectoryExists: p]]]].
	oldTime _ oldTime * 100. "Will do 100 times as many runs of the new method, so adjust time accordingly"
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[300 timesRepeat: [newTests _ paths collect: [:p | dir directoryExists: p]]]].
	^ Array
		with: ((oldTests = newTests)
			ifFalse: ['test results are not the same (this is not necessarily wrong)']
			ifTrue: ['test results are the same'])
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/12/2002 19:34'!
timeEntryAtIfAbsent: fileName inDirectoryPath: dirPath

	"self timeEntryAtIfAbsent: 'hosts' inDirectoryPath: '/etc'"
	"self timeEntryAtIfAbsent: 'SYSTEM.INI' inDirectoryPath: 'C:\WINDOWS'"

	| dir oldTime newTime |
	dir _ FileDirectory on: dirPath.
	dir entryAt: fileName.
	(dir respondsTo: #defaultEntryAt:ifAbsent:) ifFalse:
		[^ self inform: 'not used for ', Smalltalk platformName, ', no performance difference on this platform'].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[1000 timesRepeat: [dir defaultEntryAt: fileName ifAbsent: []]]].
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[1000 timesRepeat: [dir entryAt: fileName ifAbsent: []]]].
	^ Array
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/11/2002 23:54'!
timeFileAndDirectoryNamesCollection: dirPath

	"self timeFileAndDirectoryNamesCollection: '/etc'"
	"self timeFileAndDirectoryNamesCollection: 'C:\WINDOWS'"

	| dir oldTime newTime |
	"Evaluate the commands a couple of times to make sure the Unix buffer cache is
	not polluting the timing results."
	dir _ FileDirectory on: dirPath.
	dir fileAndDirectoryNames.
	dir defaultFileAndDirectoryNames.
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[1000 timesRepeat: [dir defaultFileAndDirectoryNames]]].
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[1000 timesRepeat: [dir fileAndDirectoryNames]]].
	^ Array
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/9/2002 15:32'!
timeFileExists: dirPath

	"self timeFileExists: '/etc'"
	"self timeFileExists: 'C:\WINDOWS'"

	| dir names paths oldTime oldTests newTime newTests |
	dir _ FileDirectory on: dirPath.
	names _ dir fileAndDirectoryNames.
	paths _ names collect: [:n | dirPath, FileDirectory slash, n].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[3 timesRepeat: [oldTests _ paths collect: [:p | dir defaultFileExists: p]]]].
	oldTime _ oldTime * 100. "Will do 100 times as many runs of the new method, so adjust time accordingly"
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[300 timesRepeat: [newTests _ paths collect: [:p | dir fileExists: p]]]].
	^ Array
		with: ((oldTests = newTests)
			ifFalse: ['test results are not the same (this is not necessarily wrong)']
			ifTrue: ['test results are the same'])
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/9/2002 22:56'!
timeFileOrDirectoryExists: dirPath

	"self timeFileOrDirectoryExists: '/etc'"
	"self timeFileOrDirectoryExists: 'C:\WINDOWS'"

	| dir names paths oldTime oldTests newTime newTests |
	dir _ FileDirectory on: dirPath.
	names _ dir fileAndDirectoryNames.
	paths _ names collect: [:n | dirPath, FileDirectory slash, n].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[3 timesRepeat: [oldTests _ paths collect: [:p | dir defaultFileOrDirectoryExists: p]]]].
	oldTime _ oldTime * 100. "Will do 100 times as many runs of the new method, so adjust time accordingly"
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[300 timesRepeat: [newTests _ paths collect: [:p | dir fileOrDirectoryExists: p]]]].
	^ Array
		with: ((oldTests = newTests)
			ifFalse: ['test results are not the same (this is not necessarily wrong)']
			ifTrue: ['test results are the same'])
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'performance tests' stamp: 'dtl 2/12/2002 15:17'!
timeIsAFileNamed: dirPath

	"self timeIsAFileNamed: '/etc'"
	"self timeIsAFileNamed: 'C:\WINDOWS'"

	| dir names paths oldTime oldTests newTime newTests count multiplier |
	count _ 2.
	multiplier _ 500.
	dir _ FileDirectory on: dirPath.
	names _ dir fileAndDirectoryNames.
	paths _ names collect: [:n | dirPath, FileDirectory slash, n].
	TimeProfileBrowser onBlock: [
		oldTime _ Time millisecondsToRun:
			[count timesRepeat: [oldTests _ paths collect: [:p | StandardFileStream defaultIsAFileNamed: p]]]].
	oldTime _ oldTime * multiplier. "Will do multiplier times as many runs of the new method, so adjust time accordingly"
	TimeProfileBrowser onBlock: [
		newTime _ Time millisecondsToRun:
			[(count * multiplier) timesRepeat: [newTests _ paths collect: [:p | self isAFileNamed: p]]]].
	^ Array
		with: ((oldTests = newTests)
			ifFalse: ['test results are not the same (this is not necessarily wrong)']
			ifTrue: ['test results are the same'])
		with: 'new implementation is ', (oldTime / newTime) asFloat printString, ' times faster'
		with: oldTime
		with: newTime
! !

!DirectoryPlugin class methodsFor: 'directory access' stamp: 'dtl 2/9/2002 14:41'!
getEntries: aDirectoryPath
	"Skip the first two entries if they are '.' and '..', and skip over any entry which cannot
	be read with primFileStat (this may occur if the containing directory is not executable,
	in which case it is possible to read the name of a file in the directory, but not to obtain
	its file status)."

	"self getEntries: '/etc'"
	"self getEntries: '/usr/bin'"
	"self getEntries: '/no/such/directory'"
	"self getEntries: ''"
	"self getEntries: '/'"
	"self getEntries: 'C:\WINDOWS'"

	| path names hasNil result entArray base |
	path _ self fixPath: aDirectoryPath.
	(names _ self getFileNames: path) ifNil: [^ nil].
	hasNil _ false. "This flag is set if one or more entries in the directory cannot be read with primFileStat"
	result _ names ifNotNil:
		[base _ path, FileDirectory slash.
		names collect: [:e |
			entArray _ self primStat: e inDirectoryPath: base.
			entArray isNil
				ifTrue: [hasNil _ true. nil]
				ifFalse: [DirectoryEntry fromArray: (entArray at: 1 put: e; yourself)]]].
	^ hasNil
		ifTrue: [result select: [:e | e notNil]]
		ifFalse: [result]
! !

!DirectoryPlugin class methodsFor: 'directory access' stamp: 'dtl 1/31/2002 19:01'!
getFileNames: aDirectoryPath
	"Skip the first two entries if they are '.' and '..'"

	"self getFileNames: '/etc'"
	"self getFileNames: '/usr/bin'"
	"self getFileNames: '/no/such/directory'"
	"self getFileNames: ''"
	"self getFileNames: 'C:'"
	"self getFileNames: 'C:\WINDOWS'"

	| path entName dir nameStream |
	path _ self fixPath: aDirectoryPath.
	(dir _ self primOpendir: path) ifNil: [^ nil].
	nameStream _ WriteStream on: Array new.
	entName _ self primReaddir: dir.
	entName ifNil: [^ entName].
	((entName ~= '.') and: [entName notNil]) ifTrue: [nameStream nextPut: entName].
	entName _ self primReaddir: dir.
	((entName ~= '..') and: [entName notNil]) ifTrue: [nameStream nextPut: entName].
	[(entName _ self primReaddir: dir) notNil] whileTrue: [nameStream nextPut: entName].
	self primClosedir: dir.
	^ nameStream contents

! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 17:44'!
directoryExists: pathName
	"Replacement for FileDirectory>>directoryExists"

	"self directoryExists: '/etc/hosts'"
	"self directoryExists: '/etc'"
	"self directoryExists: 'hosts'"
	"self directoryExists: 'noSuchFile'"
	"self directoryExists: '.'"
	"self directoryExists: '..'"
	"self directoryExists: '/'"
	"self directoryExists: 'C:\WINDOWS\WRITE.EXE'"
	"self directoryExists: 'C:\WINDOWS'"

	| fileExists entryArray |
	^ (fileExists _ self primFileAccessExists: pathName)
		ifNotNil:
			[fileExists and:
				[entryArray _ self primStat: pathName.
				entryArray isNil or: [entryArray at: 4]]]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 18:16'!
directoryExists: fileName inDirectoryPath: aDirectoryPath
	"Replacement for FileDirectory>>directoryExists"

	"self directoryExists: 'hosts' inDirectoryPath: '/etc/'"
	"self directoryExists: 'bin' inDirectoryPath: '/usr/'"

	"self directoryExists: 'bin' inDirectoryPath: '/usr'"
	"self directoryExists: '' inDirectoryPath: '/etc/'"
	"self directoryExists: 'hosts' inDirectoryPath: ''"
	"self directoryExists: 'noSuchFile' inDirectoryPath: ''"
	"self directoryExists: '.' inDirectoryPath: '/'"
	"self directoryExists: '..' inDirectoryPath: '/'"
	"self directoryExists: '/' inDirectoryPath: '/'"
	"self directoryExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"
	"self directoryExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS'"
	"self directoryExists: 'C:\WINDOWS' inDirectoryPath: ''"

	| fileExists entryArray |
	^ (fileExists _ self primFileAccessExists: fileName inDirectoryPath: aDirectoryPath)
		ifNotNil:
			[fileExists and:
				[entryArray _ self primStat: fileName inDirectoryPath: aDirectoryPath.
				entryArray isNil or: [entryArray at: 4]]]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 17:34'!
fileExists: pathName
	"Replacement for FileDirectory>>fileExists"

	"self fileExists: '/etc/hosts'"
	"self fileExists: '/etc'"
	"self fileExists: 'hosts'"
	"self fileExists: 'noSuchFile'"
	"self fileExists: '.'"
	"self fileExists: '..'"
	"self fileExists: '/'"
	"self fileExists: 'C:\WINDOWS\WRITE.EXE'"
	"self fileExists: 'C:\WINDOWS'"

	| fileExists entryArray |
	^ (fileExists _ self primFileAccessExists: pathName)
		ifNotNil:
			[fileExists and:
				[entryArray _ self primStat: pathName.
				(entryArray isNil or: [entryArray at: 4]) not]]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 18:17'!
fileExists: fileName inDirectoryPath: aDirectoryPath
	"Replacement for FileDirectory>>fileExists"

	"self fileExists: 'hosts' inDirectoryPath: '/etc/'"
	"self fileExists: 'bin' inDirectoryPath: '/usr/'"

	"self fileExists: 'bin' inDirectoryPath: '/usr'"
	"self fileExists: '' inDirectoryPath: '/etc/'"
	"self fileExists: 'hosts' inDirectoryPath: ''"
	"self fileExists: 'noSuchFile' inDirectoryPath: ''"
	"self fileExists: '.' inDirectoryPath: '/'"
	"self fileExists: '..' inDirectoryPath: '/'"
	"self fileExists: '/' inDirectoryPath: '/'"
	"self fileExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"
	"self fileExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS'"
	"self fileExists: 'C:\WINDOWS' inDirectoryPath: ''"

	| fileExists entryArray |
	^ (fileExists _ self primFileAccessExists: fileName inDirectoryPath: aDirectoryPath)
		ifNotNil:
			[fileExists and:
				[entryArray _ self primStat: fileName inDirectoryPath: aDirectoryPath.
				(entryArray isNil or: [entryArray at: 4]) not]]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 17:42'!
fileOrDirectoryExists: pathName
	"Replacement for FileDirectory>>fileOrDirectoryExists"

	"self fileOrDirectoryExists: '/etc/hosts'"
	"self fileOrDirectoryExists: '/etc'"
	"self fileOrDirectoryExists: 'hosts'"
	"self fileOrDirectoryExists: 'noSuchFile'"
	"self fileOrDirectoryExists: '.'"
	"self fileOrDirectoryExists: '..'"
	"self fileOrDirectoryExists: '/'"
	"self fileOrDirectoryExists: 'C:\WINDOWS\WRITE.EXE'"
	"self fileOrDirectoryExists: 'C:\WINDOWS'"

	^ self primFileAccessExists: pathName
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 18:29'!
fileOrDirectoryExists: fileName inDirectoryPath: aDirectoryPath
	"Replacement for FileDirectory>>fileOrDirectoryExists"

	"self fileOrDirectoryExists: 'hosts' inDirectoryPath: '/etc/'"
	"self fileOrDirectoryExists: 'bin' inDirectoryPath: '/usr/'"

	"self fileOrDirectoryExists: 'bin' inDirectoryPath: '/usr'"
	"self fileOrDirectoryExists: '' inDirectoryPath: '/etc/'"
	"self fileOrDirectoryExists: 'hosts' inDirectoryPath: ''"
	"self fileOrDirectoryExists: 'noSuchFile' inDirectoryPath: ''"
	"self fileOrDirectoryExists: '.' inDirectoryPath: '/'"
	"self fileOrDirectoryExists: '..' inDirectoryPath: '/'"
	"self fileOrDirectoryExists: '/' inDirectoryPath: '/'"
	"self fileOrDirectoryExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"
	"self fileOrDirectoryExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS'"
	"self fileOrDirectoryExists: 'C:\WINDOWS' inDirectoryPath: ''"

	^ self primFileAccessExists: fileName inDirectoryPath: aDirectoryPath
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/12/2002 18:17'!
getEntry: pathName

	"Do not use this method on a Win32 system. It does not answer correct file creation
	and modification times for Windows. This is because the time information returned
	in a Posix stat() call does not map to the time information in a Win32 directory entry."

	"self getEntry: '/etc/hosts'"
	"self getEntry: 'hosts'"
	"self getEntry: 'noSuchFile'"
	"self getEntry: '.'"
	"self getEntry: '..'"
	"self getEntry: '/'"
	"self getEntry: 'C:\WINDOWS\WRITE.EXE'"

	| ent |
	^ (ent _ self stat: pathName) ifNotNil: [DirectoryEntry fromArray: ent]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/12/2002 18:16'!
getEntry: fileName inDirectoryPath: aDirectoryPath 
	"Do not use this method on a Win32 system. It does not answer correct file creation
	and modification times for Windows. This is because the time information returned
	in a Posix stat() call does not map to the time information in a Win32 directory entry."

	"self getEntry: 'hosts' inDirectoryPath: '/etc'"
	"self getEntry: 'hosts' inDirectoryPath: '/etc/'"
	"self getEntry: 'hosts' inDirectoryPath: '/noSuchDir'"
	"self getEntry: 'noSuchFile' inDirectoryPath: '/etc'"
	"self getEntry: '.' inDirectoryPath: '/etc'"
	"self getEntry: '..' inDirectoryPath: '/etc'"
	"self getEntry: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS'"

	| ent |
	^ (ent _ self primStat: fileName inDirectoryPath: aDirectoryPath, FileDirectory slash)
		ifNotNil: [DirectoryEntry fromArray: ent]

! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/9/2002 17:32'!
isAFileNamed: filePathName
	"Answer true if a file of the given name exists and could be opened with
	read access. Check first to see if we have read access to filePathName, then
	make sure it is a file (not a directory)."

	"self isAFileNamed: '/etc/hosts'"
	"self isAFileNamed: '/etc/noSuchFile'"
	"self isAFileNamed: '/etc'"
	"self isAFileNamed: 'C:\WINDOWS\WRITE.EXE'"
	"self isAFileNamed: 'C:\WINDOWS'"

	| readable entryArray |
	^ (readable _ self primFileAccessReadable: filePathName)
		ifNotNil:
			[readable and: 
				[entryArray _ self primStat: filePathName.
				(entryArray isNil or: [entryArray at: 4]) not]]
! !

!DirectoryPlugin class methodsFor: 'file access' stamp: 'dtl 2/2/2002 17:49'!
stat: pathString
	"Stat a file with pathString interpreted independent of the current directoryPath.
	The sender is responsible for putting the name into the first slot of the result.
	This is to avoid the need to parse the pathString into file name and directory
	components."

	"self stat: '/etc/hosts'"
	"self stat: '/noSuchDirectory/hosts'"
	"self stat: '/etc/noSuchFile'"
	"self stat: '/usr/bin'"
	"self stat: ''"

	| a |
	^ (a _ self primStat: pathString)
		ifNotNil: [a at: 1 put: (FileDirectory localNameFor: pathString); yourself]
! !

!DirectoryPlugin class methodsFor: 'directory primitive access' stamp: 'dtl 1/30/2002 05:46'!
primClosedir: anExternalAddress
	"Close the directory stream associated with anExternalAddress. Answer
	anExternalAddress on success, or nil on failure. Caution: do not call this
	twice on the same externalAddress."

	"self primClosedir: (self primOpendir: '/etc')"
	"self primClosedir: (self primOpendir: '/no/such/directory')"

	<primitive: 'primitiveClosedir' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'directory primitive access' stamp: 'dtl 1/28/2002 22:15'!
primOpendir: pathString
	"Answer an ExternalAddress for a directory stream on pathString, or nil if
	the directory cannot be opened"

	"self primOpendir: '/etc'"
	"self primOpendir: '.'"
	"self primOpendir: '/no/such/directory'"

	<primitive: 'primitiveOpendir' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'directory primitive access' stamp: 'dtl 1/30/2002 18:51'!
primReaddir: anExternalAddress
	"Read the next directory entry from the directory stream associated with
	anExternalAddress. Answer and array suitable for creating a DirectoryEntry,
	or nil for an error or end of directory stream."

	"self primReaddir: (self primOpendir: '/etc')"
	"self primReaddir: (self primOpendir: '/no/such/directory')"

	<primitive: 'primitiveReaddir' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'directory primitive access' stamp: 'dtl 1/31/2002 17:11'!
primRewinddir: anExternalAddress
	"Rewind the directory stream associated with anExternalAddress. Answer
	anExternalAddress on success, or nil on failure."

	"self primRewinddir: (self primOpendir: '/etc')"
	"self primRewinddir: (self primOpendir: '/no/such/directory')"

	<primitive: 'primitiveRewinddir' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'directory primitive access' stamp: 'dtl 1/27/2002 10:47'!
primSetDirectory: pathString

	"self primSetDirectory: '/etc'"
	"self primSetDirectory: '.'"
	"self primSetDirectory: '/no/such/directory'"

	<primitive: 'primitiveSetDirectory' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:20'!
primFileAccessExecutable: pathString
	"Check if file is executable by this OS process."

	"self primFileAccessExecutable: '/etc/hosts'"
	"self primFileAccessExecutable: '/noSuchDirectory/hosts'"
	"self primFileAccessExecutable: '/etc/noSuchFile'"
	"self primFileAccessExecutable: '/usr/bin'"
	"self primFileAccessExecutable: '/bin/sh'"
	"self primFileAccessExecutable: ''"
	"self primFileAccessExecutable: 'C:\WINDOWS\WRITE.EXE'"

	<primitive: 'primitiveFileAccessExecutable' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:38'!
primFileAccessExecutable: fileName inDirectoryPath: directoryPath
	"Check if file is executable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	"self primFileAccessExecutable: 'hosts' inDirectoryPath: '/etc/'"
	"self primFileAccessExecutable: 'noSuchFile' inDirectoryPath: '/etc/'"
	"self primFileAccessExecutable: 'hosts' inDirectoryPath: '/no/such/directory/'"
	"self primFileAccessExecutable: 'hosts' inDirectoryPath: '/etc'"	"Bad path, should have a slash"
	"self primFileAccessExecutable: 'bin' inDirectoryPath: '/usr/'"
	"self primFileAccessExecutable: '' inDirectoryPath: ''"
	"self primFileAccessExecutable: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"

	<primitive: 'primitiveFileAccessExecutableInDirectory' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:59'!
primFileAccessExists: pathString
	"Check for existence of a file with a call to access()."

	"self primFileAccessExists: '/etc/hosts'"
	"self primFileAccessExists: '/noSuchDirectory/hosts'"
	"self primFileAccessExists: '/etc/noSuchFile'"
	"self primFileAccessExists: '/usr/bin'"
	"self primFileAccessExists: '/bin/sh'"
	"self primFileAccessExists: ''"
	"self primFileAccessExists: 'C:\WINDOWS\WRITE.EXE'"
	"self primFileAccessExists: 'C:\WINDOWS'"

	<primitive: 'primitiveFileAccessExists' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:40'!
primFileAccessExists: fileName inDirectoryPath: directoryPath
	"Check for existence of a file in a directory with a call to access(). The
	directory path string is expected to include a trailing directory separator,
	for example '/etc/' rather than '/etc'."

	"self primFileAccessExists: 'hosts' inDirectoryPath: '/etc/'"
	"self primFileAccessExists: 'noSuchFile' inDirectoryPath: '/etc/'"
	"self primFileAccessExists: 'hosts' inDirectoryPath: '/no/such/directory/'"
	"self primFileAccessExists: 'hosts' inDirectoryPath: '/etc'"	"Bad path, should have a slash"
	"self primFileAccessExists: 'bin' inDirectoryPath: '/usr/'"
	"self primFileAccessExists: '' inDirectoryPath: ''"
	"self primFileAccessExists: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"

	<primitive: 'primitiveFileAccessExistsInDirectory' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:29'!
primFileAccessReadable: pathString
	"Check if file is readable by this OS process."

	"self primFileAccessReadable: '/etc/hosts'"
	"self primFileAccessReadable: '/noSuchDirectory/hosts'"
	"self primFileAccessReadable: '/etc/noSuchFile'"
	"self primFileAccessReadable: '/usr/bin'"
	"self primFileAccessReadable: '/bin/sh'"
	"self primFileAccessReadable: ''"
	"self primFileAccessReadable: 'C:\WINDOWS\WRITE.EXE'"

	<primitive: 'primitiveFileAccessReadable' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:37'!
primFileAccessReadable: fileName inDirectoryPath: directoryPath
	"Check if file is readable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	"self primFileAccessReadable: 'hosts' inDirectoryPath: '/etc/'"
	"self primFileAccessReadable: 'noSuchFile' inDirectoryPath: '/etc/'"
	"self primFileAccessReadable: 'hosts' inDirectoryPath: '/no/such/directory/'"
	"self primFileAccessReadable: 'hosts' inDirectoryPath: '/etc'"	"Bad path, should have a slash"
	"self primFileAccessReadable: 'bin' inDirectoryPath: '/usr/'"
	"self primFileAccessReadable: '' inDirectoryPath: ''"
	"self primFileAccessReadable: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"

	<primitive: 'primitiveFileAccessReadableInDirectory' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:29'!
primFileAccessWritable: pathString
	"Check if file is writable by this OS process."

	"self primFileAccessWritable: '/etc/hosts'"
	"self primFileAccessWritable: '/noSuchDirectory/hosts'"
	"self primFileAccessWritable: '/etc/noSuchFile'"
	"self primFileAccessWritable: '/usr/bin'"
	"self primFileAccessWritable: '/bin/sh'"
	"self primFileAccessWritable: ''"
	"self primFileAccessWritable: 'C:\WINDOWS\WRITE.EXE'"

	<primitive: 'primitiveFileAccessWritable' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/9/2002 14:39'!
primFileAccessWritable: fileName inDirectoryPath: directoryPath
	"Check if file is writable by this OS process. The directory path string is expected
	to include a trailing directory separator, for example '/etc/' rather than '/etc'."

	"self primFileAccessWritable: 'hosts' inDirectoryPath: '/etc/'"
	"self primFileAccessWritable: 'noSuchFile' inDirectoryPath: '/etc/'"
	"self primFileAccessWritable: 'hosts' inDirectoryPath: '/no/such/directory/'"
	"self primFileAccessWritable: 'hosts' inDirectoryPath: '/etc'"	"Bad path, should have a slash"
	"self primFileAccessWritable: 'bin' inDirectoryPath: '/usr/'"
	"self primFileAccessWritable: '' inDirectoryPath: ''"
	"self primFileAccessWritable: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"

	<primitive: 'primitiveFileAccessWritableInDirectory' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/10/2002 21:11'!
primStat: pathString
	"Stat a file with named by pathString, a directory path and file name.
	Timezone offset calculation is done in the primitive; compare to #primStatArray:.
	The sender is responsible for putting the name into the first slot of the result.
	This is to avoid the need to parse the pathString into file name and directory
	components."

	"self primStat: '/etc/hosts'"
	"self primStat: '/noSuchDirectory/hosts'"
	"self primStat: '/etc/noSuchFile'"
	"self primStat: '/usr/bin'"
	"self primStat: ''"
	"self primStat: 'C:\WINDOWS\WRITE.EXE'"

	<primitive: 'primitiveStat' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'file primitive access' stamp: 'dtl 2/10/2002 21:11'!
primStat: fileName inDirectoryPath: aSlashTerminatedDirectoryPath
	"Stat a file named fileName located in aSlashTerminatedDirectoryPath. Timezone
	offset calculation is done in the primitive; compare to #primStatArray:inDirectoryPath."

	"self primStat: 'hosts' inDirectoryPath: '/etc/'"
	"self primStat: 'noSuchFile' inDirectoryPath: '/etc/'"
	"self primStat: 'hosts' inDirectoryPath: '/no/such/directory/'"
	"self primStat: 'hosts' inDirectoryPath: '/etc'"	"This is a bad path, should have a slash"
	"self primStat: 'bin' inDirectoryPath: '/usr/'"
	"self primStat: '' inDirectoryPath: ''"
	"self primStat: 'WRITE.EXE' inDirectoryPath: 'C:\WINDOWS\'"

	<primitive: 'primitiveStatInPath' module: 'DirectoryPlugin'>
	^ nil
! !

!DirectoryPlugin class methodsFor: 'other primitive access' stamp: 'dtl 1/27/2002 10:47'!
primModuleIsLoaded

	"self primModuleIsLoaded"

	<primitive: 'primitiveModuleIsLoaded' module: 'DirectoryPlugin'>
	^ false
! !

!DirectoryPlugin class methodsFor: 'private' stamp: 'dtl 1/31/2002 19:00'!
fixPath: pathString
	"An empty path string is assumed to refer to the current directory. opendir()
	on an empty string will fail, so give it an explicit path to the current
	directory."

	^ pathString isEmpty
		ifTrue: ['.']
		ifFalse: [pathString]! !


!FileDirectory methodsFor: 'enumeration' stamp: 'dtl 1/27/2002 15:30'!
defaultFileAndDirectoryNames
	"FileDirectory default defaultFileAndDirectoryNames"

	^ self entries collect: [:entry | entry first]
! !

!FileDirectory methodsFor: 'enumeration' stamp: 'dtl 1/27/2002 15:32'!
fileAndDirectoryNames

	"FileDirectory default fileAndDirectoryNames"

	^ ((Smalltalk
		at: #DirectoryPlugin
		ifAbsent: [^ self defaultFileAndDirectoryNames])
		getFileNames: pathName)
		ifNil: [^ self defaultFileAndDirectoryNames]
! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 8/4/2003 06:56'!
defaultDirectoryExists: filenameOrPath
	"Answer true if a directory of the given name exists. The given name may be either a full path name or a local directory within this directory."
	"FileDirectory default directoryExists: FileDirectory default pathName"

	| fName dir |
	FileDirectory activeDirectoryClass splitName: filenameOrPath to:
		[:filePath :name |
			fName _ name.
			filePath isEmpty
				ifTrue: [dir _ self]
				ifFalse: [dir _ FileDirectory on: filePath]].

	^dir exists and: [
		self isCaseSensitive 
			ifTrue:[dir directoryNames includes: fName]
			ifFalse:[dir directoryNames anySatisfy: [:name| name sameAs: fName]]].
! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 8/4/2003 07:03'!
defaultFileExists: filenameOrPath
	"Answer true if a file of the given name exists. The given name may be either a full path name or a local file within this directory."
	"FileDirectory default fileExists: Smalltalk sourcesName"

	| fName dir |
	FileDirectory activeDirectoryClass splitName: filenameOrPath to:
		[:filePath :name |
			fName _ name.
			filePath isEmpty
				ifTrue: [dir _ self]
				ifFalse: [dir _ FileDirectory on: filePath]].
	self isCaseSensitive 
		ifTrue:[^dir fileNames includes: fName]
		ifFalse:[^dir fileNames anySatisfy: [:name| name sameAs: fName]].	! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 8/4/2003 07:05'!
defaultFileOrDirectoryExists: filenameOrPath
	"Answer true if either a file or a directory file of the given name exists. The given name may be either a full path name or a local name within this directory."
	"FileDirectory default fileOrDirectoryExists: Smalltalk sourcesName"

	| fName dir |
	FileDirectory activeDirectoryClass splitName: filenameOrPath to:
		[:filePath :name |
			fName _ name.
			filePath isEmpty
				ifTrue: [dir _ self]
				ifFalse: [dir _ FileDirectory on: filePath]].

	^ (dir includesKey: fName) or: [ fName = '' and:[ dir entries size > 1]]! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 2/9/2002 18:38'!
directoryExists: filenameOrPath
	"Answer true if a directory of the given name exists. The given name may be either a full path name or a local directory within this directory."

	"FileDirectory default directoryExists: FileDirectory default pathName"

	| pluginClass |
	pluginClass _ Smalltalk at: #DirectoryPlugin
		ifAbsent: [^ self defaultDirectoryExists: filenameOrPath].
	^ (self class splitName: filenameOrPath to:
		[:filePath :name |
			filePath isEmpty
				ifTrue: [pluginClass directoryExists: filenameOrPath
									inDirectoryPath: self pathName, FileDirectory slash]
				ifFalse: [pluginClass directoryExists: filenameOrPath]])
		ifNil: [self defaultDirectoryExists: filenameOrPath]
! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 2/9/2002 18:37'!
fileExists: filenameOrPath
	"Answer true if a file of the given name exists. The given name may be either a full path name or a local file within this directory."

	"FileDirectory default fileExists: Smalltalk sourcesName"

	| pluginClass |
	pluginClass _ Smalltalk at: #DirectoryPlugin
		ifAbsent: [^ self defaultFileExists: filenameOrPath].
	^ (self class splitName: filenameOrPath to:
		[:filePath :name |
			filePath isEmpty
				ifTrue: [pluginClass fileExists: filenameOrPath
									inDirectoryPath: self pathName, FileDirectory slash]
				ifFalse: [pluginClass fileExists: filenameOrPath]])
		ifNil: [self defaultFileExists: filenameOrPath]
! !

!FileDirectory methodsFor: 'testing' stamp: 'dtl 2/9/2002 18:39'!
fileOrDirectoryExists: filenameOrPath
	"Answer true if either a file or a directory file of the given name exists. The given name may be either a full path name or a local name within this directory."
	"FileDirectory default fileOrDirectoryExists: Smalltalk sourcesName"

	| pluginClass |
	pluginClass _ Smalltalk at: #DirectoryPlugin
		ifAbsent: [^ self defaultFileOrDirectoryExists: filenameOrPath].
	^ (self class splitName: filenameOrPath to:
		[:filePath :name |
			filePath isEmpty
				ifTrue: [pluginClass fileOrDirectoryExists: filenameOrPath
									inDirectoryPath: self pathName, FileDirectory slash]
				ifFalse: [pluginClass fileOrDirectoryExists: filenameOrPath]])
		ifNil: [self defaultFileOrDirectoryExists: filenameOrPath]
! !

!FileDirectory methodsFor: 'private' stamp: 'ar 5/30/2001 20:49'!
directoryContentsFor: fullPath
	"Return a collection of directory entries for the files and directories in the directory with the given path. See primLookupEntryIn:index: for further details."
	"FileDirectory default directoryContentsFor: ''"

	| entries index done entryArray |
	entries _ OrderedCollection new: 200.
	index _ 1.
	done _ false.
	[done] whileFalse: [
		entryArray _ self primLookupEntryIn: fullPath index: index.
		#badDirectoryPath = entryArray ifTrue: [
			^(InvalidDirectoryError pathName: pathName) signal].
		entryArray == nil
			ifTrue: [done _ true]
			ifFalse: [entries addLast: (DirectoryEntry fromArray: entryArray)].
		index _ index + 1].

	^ entries asArray
! !

!FileDirectory methodsFor: 'file status' stamp: 'sma 6/3/2000 22:11'!
entryAt: fileName ifAbsent: aBlock
	"Find the entry with local name fileName and answer it.
	If not found, answer the result of evaluating aBlock."

	| comparisonBlock |
	self isCaseSensitive
		ifTrue: [comparisonBlock _ [:entry | (entry at: 1) = fileName]]
		ifFalse: [comparisonBlock _ [:entry | (entry at: 1) sameAs: fileName]].
	^ self entries detect: comparisonBlock ifNone: [aBlock value]! !


!StandardFileStream class methodsFor: 'file creation' stamp: 'dtl 2/9/2002 15:28'!
defaultIsAFileNamed: fileName
	"Answer true if a file of the given name exists."

	| f |
	f _ self new open: fileName forWrite: false.
	f ifNil: [^ false].
	f close.
	^ true
! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'dtl 2/9/2002 15:29'!
isAFileNamed: fileName 
	"Answer true if a file of the given name exists. If UnixDirectoryPlugin 
	is available, use it. Otherwise fall back on the portable mechanism."

	^ ((Smalltalk
		at: #DirectoryPlugin
		ifAbsent: [^ self defaultIsAFileNamed: fileName])
		isAFileNamed: fileName)
		ifNil: [^ self defaultIsAFileNamed: fileName]! !


!UnixFileDirectory methodsFor: 'private' stamp: 'dtl 2/10/2002 13:58'!
defaultDirectoryContentsFor: fullPath
	"This method exists to support timing comparisons between my implementation of
	#directoryContentsFor and the default implementation."

	^ super directoryContentsFor: fullPath! !

!UnixFileDirectory methodsFor: 'private' stamp: 'dtl 1/27/2002 17:56'!
directoryContentsFor: fullPath
	"Return a collection of directory entries for the files and directories in the directory
	with the given path. If UnixDirectoryPlugin is available, use it. Otherwise fall back
	on the slower method in FileDirectory."

	"FileDirectory default directoryContentsFor: ''"

	^ ((Smalltalk
		at: #DirectoryPlugin
		ifAbsent: [^ super directoryContentsFor: fullPath])
		getEntries: fullPath)
		ifNil: [^ super directoryContentsFor: fullPath]
! !

!UnixFileDirectory methodsFor: 'file status' stamp: 'dtl 2/10/2002 13:08'!
defaultEntryAt: fileName ifAbsent: aBlock
	"This method exists to support timing comparisons between my implementation of
	#entryAt:ifAbsent and the default implementation."

	^ super entryAt: fileName ifAbsent: aBlock! !

!UnixFileDirectory methodsFor: 'file status' stamp: 'dtl 2/10/2002 13:03'!
entryAt: fileName ifAbsent: aBlock
	"Find the entry with local name fileName and answer it.
	If not found, answer the result of evaluating aBlock."

	"(UnixFileDirectory on: '/etc') entryAt: 'hosts' ifAbsent: ['no such file']"
	"(UnixFileDirectory on: '/etc') entryAt: 'noSuchFile' ifAbsent: ['no such file']"
	"(UnixFileDirectory on: '/noSuchDirectory') entryAt: 'hosts' ifAbsent: ['no such file']"


	(Smalltalk
		at: #DirectoryPlugin
		ifAbsent: [^ super entryAt: fileName ifAbsent: aBlock])
		primModuleIsLoaded
			ifTrue: [^ (DirectoryPlugin getEntry: fileName inDirectoryPath: self pathName)
						ifNil: [aBlock value]]
			ifFalse: [^ super entryAt: fileName ifAbsent: aBlock]
! !


!DirectoryPlugin class reorganize!
('translation' declareCVarsIn: moduleName sdeclareCVarsIn: shouldBeTranslated)
('version testing' installedModuleVersionString versionString)
('performance tests' compareIsAFileNamed: timeDirectoryEntryCollection: timeDirectoryExists: timeEntryAtIfAbsent:inDirectoryPath: timeFileAndDirectoryNamesCollection: timeFileExists: timeFileOrDirectoryExists: timeIsAFileNamed:)
('directory access' getEntries: getFileNames:)
('file access' directoryExists: directoryExists:inDirectoryPath: fileExists: fileExists:inDirectoryPath: fileOrDirectoryExists: fileOrDirectoryExists:inDirectoryPath: getEntry: getEntry:inDirectoryPath: isAFileNamed: stat:)
('directory primitive access' primClosedir: primOpendir: primReaddir: primRewinddir: primSetDirectory:)
('file primitive access' primFileAccessExecutable: primFileAccessExecutable:inDirectoryPath: primFileAccessExists: primFileAccessExists:inDirectoryPath: primFileAccessReadable: primFileAccessReadable:inDirectoryPath: primFileAccessWritable: primFileAccessWritable:inDirectoryPath: primStat: primStat:inDirectoryPath:)
('other primitive access' primModuleIsLoaded)
('private' fixPath:)
!


!DirectoryPlugin reorganize!
('directory primitives' primitiveClosedir primitiveOpendir primitiveReaddir primitiveRewinddir)
('file primitives' primitiveFileAccessExecutable primitiveFileAccessExecutableInDirectory primitiveFileAccessExists primitiveFileAccessExistsInDirectory primitiveFileAccessReadable primitiveFileAccessReadableInDirectory primitiveFileAccessWritable primitiveFileAccessWritableInDirectory primitiveStat primitiveStatInPath)
('other primitives' primitiveModuleIsLoaded primitiveVersionString)
('version string' versionString)
('private - directory' openDirectoryStream:)
('private - file' checkAccess:directoryPath:mode: checkAccess:mode: entryArrayFor:fromStatBuf: fileExecutableFlag fileOKFlag fileReadableFlag fileWritableFlag makeDirEntryName:createDate:modDate:isDir:fileSize: putStatFor:intoBuffer: statFile:intoBuffer: statFile:intoBuffer:directoryPath: statFileToEntryArray: statToEntryArray:directoryPath:)
('private - security' canOpenDirectoryStreamFor:length: canStatFilePath:length:)
('private - timezone' smalltalkSecondsFromPosixSeconds:)
('private - misc' addressObjectFor: pointerFrom: sizeOfPointer smallOrLargeIntegerObjectFor: stringFromCString:)
!

